@@ -1,29 +1,36 @@
 import datetime as dt
-from delta.tables import DeltaTable
 import json
-from odapdbxbundle.common.databricks import resolve_dbutils
-from odapdbxbundle.common.logger import logger
+from urllib import parse
+import uuid
+import requests
+import yaml
+from delta.tables import DeltaTable
 from pyspark.sql import SparkSession, DataFrame
 from pyspark.sql import SQLContext
 import pyspark.sql.functions as F
 import pyspark.sql.types as T
-from pyspark.sql.types import *
-import requests
-from urllib import parse
-import uuid
-import yaml
 from kbcstorage.files import Files
 from kbcstorage.tables import Tables
 from kbcstorage.workspaces import Workspaces
+from odapdbxbundle.common.databricks import resolve_dbutils
+from odapdbxbundle.common.logger import logger
 
 SECONDS_IN_HOUR = 3600
 
 
+class NonexistingTable(Exception):
+    pass
+
+
+class MissingSchema(Exception):
+    pass
+
+
 class KeboolaExporter:
     def __init__(self, config_path):
         self._spark = SparkSession.getActiveSession()
 
-        with open(config_path, "r") as yamlfile:
+        with open(config_path, "r", encoding="utf8") as yamlfile:
             self.config = yaml.load(yamlfile, Loader=yaml.FullLoader)
         self._dbutils = resolve_dbutils()
         self._kbc_token = self._dbutils.secrets.get(self.config["scope"], self.config["secret"])
@@ -68,7 +75,7 @@ class KeboolaExporter:
             workspace = self.__workspace_connection(workspaces_api, workspace_id)
             container_extract, url_extract, sas = self.__extract_workspace_connection_detail(workspace)
             assert (
-                container == container_extract and url_extract == f"{account}.{windows_url_service}"
+                    container == container_extract and url_extract == f"{account}.{windows_url_service}"
             ), f"Workspace detail does not match expected workspace container: {container_extract} url: {url_extract}."
 
             self._spark.conf.set(f"fs.azure.sas.{container_extract}.{url_extract}", sas)
@@ -87,24 +94,25 @@ class KeboolaExporter:
         return workspace
 
     @staticmethod
-    def __add_schema_column(schema: StructType, name: str, column_type: str, doubles: list):
+    def __add_schema_column(schema: T.StructType, name: str, column_type: str, doubles: list):
         """
         Add new column definition to StructType based on string name of the type.
         Storing doubles and booleans columns for conversion from string.
         """
 
         type_mapping = {
-            "CHAR": StringType(),
-            "INT": LongType(),
-            "DATE": DateType(),
-            "FLOAT": StringType(),
-            "DOUBLE": StringType(),
-            "NUMERIC": StringType(),
-            "DECIMAL": StringType(),
-            "BIT": ShortType(),
+            "CHAR": T.StringType(),
+            "INT": T.LongType(),
+            "TIME": T.TimestampType(),
+            "DATE": T.DateType(),
+            "FLOAT": T.StringType(),
+            "DOUBLE": T.StringType(),
+            "NUMERIC": T.StringType(),
+            "DECIMAL": T.StringType(),
+            "BIT": T.ShortType(),
         }
 
-        spark_type = StringType()
+        spark_type = T.StringType()
         for k, v in type_mapping.items():
             if k in column_type:
                 spark_type = v
@@ -116,23 +124,27 @@ class KeboolaExporter:
         return schema
 
     def __get_keboola_schema(self, table_id: str, kbc_token: str):
-        r = requests.get(self.config["kbc_api"], headers={"X-StorageApi-Token": kbc_token})
-        parsed = json.loads(r.content)
+        response = requests.get(self.config["kbc_api"], headers={"X-StorageApi-Token": kbc_token}, timeout=30)
+        parsed = json.loads(response.content)
         for table in parsed:
             tbl_id = table["id"]
             if tbl_id == table_id:
                 return table
-        return {}
+        raise NonexistingTable(f"Table {table_id} does not exist.")
 
-    def get_schema(self, table_id: str, kbc_token: str):
+    def get_schema(self, table_id: str, kbc_token: str, scope: str):
         doubles = []
-        kbc_token_src = self._dbutils.secrets.get(self.config["scope"], kbc_token)
+        kbc_token_src = self._dbutils.secrets.get(scope, kbc_token)
         table = self.__get_keboola_schema(table_id, kbc_token_src)
-        schema = StructType()
+        schema = T.StructType()
         columns_metadata = (
-            table["columnMetadata"] if table["columnMetadata"] else table["sourceTable"]["columnMetadata"]
+            table["columnMetadata"] if "columnMetadata" in table and table["columnMetadata"]
+            else table["sourceTable"]["columnMetadata"]
         )
 
+        if not columns_metadata:
+            raise MissingSchema(f"Table {table_id} does not have schema.")
+
         for column in columns_metadata.keys():
             for column_type in columns_metadata[column]:
                 if column_type["key"] == "KBC.datatype.type" and column not in schema.names:
@@ -145,7 +157,7 @@ class KeboolaExporter:
         return table_schema, doubles
 
     def __get_changed_since(self, table_path: str):
-        if self._spark._jsparkSession.catalog().tableExists(table_path):
+        if self._spark._jsparkSession.catalog().tableExists(table_path):  # pylint: disable=[protected-access]
             return self._spark.sql(f"SHOW TBLPROPERTIES {table_path}('lastChange')").collect()[0]["value"]
         return None
 
@@ -162,7 +174,7 @@ class KeboolaExporter:
         file_path = f"wasbs://{container}@{account_core}/{filename}"
         return file_path
 
-    def __get_table_as_df(self, file_path: str, schema: StructType, doubles: list):
+    def __get_table_as_df(self, file_path: str, schema: T.StructType, doubles: list):
         chunks = [f.path for f in self._dbutils.fs.ls(file_path) if f.size != 0]
         df = (
             self._spark.read.schema(schema)
@@ -182,10 +194,11 @@ class KeboolaExporter:
         self._spark.sql(f"CREATE DATABASE IF NOT EXISTS {catalog_name}.{database_name}")
 
         # if table already exists and there are primary keys then upsert otherwise append
-        if len(table_schema["primaryKey"]) == 0 or not self._spark._jsparkSession.catalog().tableExists(table_path):
+        if len(table_schema["primaryKey"]) == 0 or not self._spark._jsparkSession.catalog().tableExists(
+                table_path):  # pylint: disable=[protected-access]
             self.__append(df, table_path, partition_by)
         else:
-            self.__upsert(df, table_path, table_schema, partition_by)
+            self.__upsert(df, table_path, table_schema)
 
         # save information about last change to table properties
         last_change = table_schema["lastChange"]
@@ -201,27 +214,20 @@ class KeboolaExporter:
 
         df_writer.saveAsTable(table_path)
 
-    def __upsert(self, df: DataFrame, table_path: str, table_schema: dict, partition_by: str):
+    def __upsert(self, df: DataFrame, table_path: str, table_schema: dict):
         primary_key = table_schema["primaryKey"]
         merge_cond = " AND ".join(f"oldData.{pk} = newData.{pk}" for pk in primary_key)
 
-        if partition_by:
-            df_partitions = self._spark.sql(f"show partitions {table_path}").toPandas()
-            partitions = df_partitions["partition"].str.replace("dt=", "").tolist()
-            partition_cond = f"{partition_by} IN ({' '.join(map(str, partitions))}) "
-            merge_cond = partition_cond + "AND " + merge_cond
-
         insert_set = {col: f"newData.`{col}`" for col in df.columns}
         update_set = {col: f"newData.`{col}`" for col in df.columns if col not in table_schema["primaryKey"]}
 
         old_df = DeltaTable.forName(self._spark, table_path)
 
-        source_columns = df.columns
-        target_columns = old_df.toDF().columns
-        missing_columns = list(set(source_columns) - set(target_columns))
+        missing_columns = list(set(df.columns) - set(old_df.toDF().columns))
         # Add the missing columns to the target dataframe, filling in with default values
         for col in missing_columns:
             self._spark.sql(f"ALTER TABLE {table_path} ADD COLUMNS ({col} string)")
+
         (
             old_df.alias("oldData")
             .merge(df.alias("newData"), merge_cond)
@@ -230,14 +236,16 @@ class KeboolaExporter:
             .execute()
         )
 
-    def export_table(self, table_id: str, table_path: str, token: str, partition_by=""):
+    def export_table(self, table_id: str, table_path: str, token: str, scope=None, partition_by=""):
         retries = 3
+        if scope is None:
+            scope = self.config["scope"]
         while retries:
             try:
-                kbc_token_src = self._dbutils.secrets.get(self.config["scope"], token)
+                kbc_token_src = self._dbutils.secrets.get(scope, token)
                 files_src = Files(self.config["kbc_url"], kbc_token_src)
                 tables_src = Tables(self.config["kbc_url"], kbc_token_src)
-                table_schema, doubles = self.get_schema(table_id, token)
+                table_schema, doubles = self.get_schema(table_id, token, scope)
                 last_change = table_schema["lastChange"]
 
                 file_id = tables_src.export(
@@ -249,28 +257,32 @@ class KeboolaExporter:
                 )
 
                 file_path = self.__get_path_to_exported_table(files_src, file_id)
-                schema = table_schema["schema"]
-                df = self.__get_table_as_df(file_path, schema, doubles)
-                rows_num = df.count()
-                logger.info(f"Exporting table: {table_id}, Last update: {last_change},  Rows: {rows_num}")
+                df = self.__get_table_as_df(file_path, table_schema["schema"], doubles)
+                logger.info(f"Exporting table: {table_id}, Last update: {last_change},  Rows: {df.count()}")
 
-                if rows_num > 0:
-                    if "DELETED_FLAG" in df.columns:
-                        df = df.filter(F.col("DELETED_FLAG") == "N")
+                if df.count() > 0:
                     self.__store_table_to_database(df, table_schema, table_path, partition_by)
-
-                self.__write_export_log(table_id, table_path, rows_num)
+                    if "DELETED_FLAG" in df.columns:
+                        new_df = DeltaTable.forName(self._spark, table_path)
+                        new_df.delete(F.col("DELETED_FLAG") == "Y")
+                        new_df.vacuum()
+                self.__write_export_log(table_id, table_path, df.count())
                 return self._spark.read.table(table_path)
             except requests.ConnectionError:
                 self._keboola_workspace_path = self.__initialize_workspace_connection(self.config)
                 retries -= 1
 
     def export_tables(self, config_path: str):
-        with open(config_path, "r") as yamlfile:
+        with open(config_path, "r", encoding="utf8") as yamlfile:
             tables = yaml.load(yamlfile, Loader=yaml.FullLoader)
 
         for table in tables["tables"]:
-            self.export_table(table["table_id"], table["table_path"], table["token"], table["partition_by"])
+            try:
+                self.export_table(table["table_id"], table["table_path"], table["token"], table["partition_by"])
+            except MissingSchema:
+                logger.info(f"Missing schema in table {table['name']}")
+            except NonexistingTable:
+                logger.info(f"Table {table['name']} does not exist.")
 
     def __write_export_log(self, table_id: str, path: str, rows: int):
         spark = SparkSession.getActiveSession()
